""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import json
from time import sleep
from subprocess import Popen, PIPE
import threading
import sys
import schedule
import re
import shlex

from sqlalchemy.exc import DBAPIError

sys.path.append("/opt/cyops-auth")

#Include common packages in ha service
sys.path.append("/opt/cyops/python_packages/fortisoar")
from handlerworkers import cluster
from utilities.csconstants import CSConstants
from utilities.csconstants import FILE_HANDLER, MESSAGE_HANDLER

from log.cslog import get_ha_logger
from utilities.ha.services import get_cyops_services_list
from utilities.ha import handler
from utilities.ha.consts import COMMAND_ID_REFRESH_CONFIG
from handlerworkers.apikeys import CSAPIKeys
from handlerworkers.publisher import publisher
from utilities.ha.supported_commands import LIST_OF_SUPPORTED_CMDS
import requests
from utilities.ha.common_utils import execute_cmd
from utilities.ha.node import get_node_from_db_or_list_node_file
from handlerworkers.cluster import save_file_from_db_to_disk, cleanup_shared_data
from utilities.ha.commands.files_handler import execute_file_handler_copy_cmd, CLUSTER_WORKSPACE_PATH, \
    execute_file_handler_delete_cmd, execute_file_handler_copy_common_files, execute_file_handler_rename_cmd

logger = get_ha_logger(__name__)

# Common response for all cluster commands. Handles both API and CLI response
def build_cluster_command_response(exit_code=0, status_code=200, output="", error=""):
    return {
        "exit_code": exit_code,
        "status_code": status_code,
        "output": output,
        "error": error
    }

def handle_regenerate_appliance_keys_command(uuid):
    CSAPIKeys().save_api_keys_to_files(uuid)
    return 0, '', ''

def check_if_supported_cmd(command):
    is_valid = False
    command = command.strip()
    for cmd_regex in LIST_OF_SUPPORTED_CMDS:
        if re.search(cmd_regex, command):
            is_valid = True
            break
    return is_valid

def api_call_handler(api_call_details):
    method = api_call_details['method']
    headers = api_call_details.get('headers')
    if not headers:
        headers = {}
    headers["is-cluster-call"] = "True"
    body = api_call_details.get('body', {})
    url = api_call_details['url']
    verify_cert = api_call_details.get('verify_cert', True)

    try:
        if method == 'DELETE' or method == 'GET':
            # When content-lenght header is present, the delete request fails
            headers.pop('content-length', None)
            response_obj = requests.request(method, url, headers=headers, verify=verify_cert)
        else:
            body = body if type(body) is str else json.dumps(body)
            response_obj = requests.request(method, url, data=body, headers=headers, verify=verify_cert)
        if response_obj.status_code != 200:
            return build_cluster_command_response(exit_code=1, status_code=response_obj.status_code)
        else:
            output=""
            if response_obj.text:
                output = response_obj.json()
            return build_cluster_command_response(exit_code=0, status_code=response_obj.status_code, output=output)
    except Exception as error:
        logger.exception(error)
        logger.error("API failed with error: "+str(error))
        return build_cluster_command_response(exit_code=1, status_code=500, error=str(error))

# message_handler will replicate a rabbitmq message across cluster
def message_handler(command_type, payload):
    try:
        message = payload.get('message')
        routing_key = payload.get('routing_key')
        exchange = payload.get('exchange')
        try:
            publisher.publish(json.loads(message), routing_key, exchange)
        except TypeError:
            publisher.publish(message, routing_key, exchange)
        return build_cluster_command_response(exit_code=0, status_code=200)
    except Exception as e:
        logger.exception('Error processing messages to publish')
        return build_cluster_command_response(exit_code=1, status_code=500)

def cmd_handler(command_type, payload):
    # for backward compatibility, command_type itself can contain the actual command to run.
    # command_type should be used as a short description for the command being run.
    command = command_type
    if payload:
        command = payload.get('command', command_type)
    if not check_if_supported_cmd(command):
        error = "Command [" + command + "] is not whitelisted in framework"
        logger.warning('Command failed with error: ' + error)
        return build_cluster_command_response(exit_code=1, error=str(error))
    command = shlex.split(command)
    logger.debug('Executing command ' + str(command)[:150])

    command_result = Popen(command, stdout=PIPE, stderr=PIPE)
    output, error = command_result.communicate()
    output = output.strip().decode('utf-8')
    error = error.strip().decode('utf-8')
    returncode = command_result.returncode
    return build_cluster_command_response(exit_code=returncode, output=output, error=error)

def file_handler(command_type, payload):
    if command_type == 'copy':
        type_identifier_in_cluster_shared_data = payload['type_identifier_in_cluster_shared_data']
        tar_file_location = CLUSTER_WORKSPACE_PATH+str(type_identifier_in_cluster_shared_data)+".tar"
        save_file_from_db_to_disk(type_identifier_in_cluster_shared_data, tar_file_location)
        exit_code, message = execute_file_handler_copy_cmd(payload)
        return build_cluster_command_response(exit_code= exit_code, output=message, error=message)
    if command_type == 'delete':
        exit_code, message = execute_file_handler_delete_cmd(payload)
        return build_cluster_command_response(exit_code=exit_code, output=message, error=message)
    if command_type == 'rename':
        exit_code, message = execute_file_handler_rename_cmd(payload)
        return build_cluster_command_response(exit_code=exit_code, output=message, error=message)
    if command_type == 'copy_common_files':
        exit_code, message  = execute_file_handler_copy_common_files(payload)
        return build_cluster_command_response(exit_code=exit_code, output=message, error=message)
    return build_cluster_command_response(exit_code=1, status_code=500, error="Invalid file command")

def run_cluster_command_new(pending_command):
    try:
        if pending_command.handler_type == "api-call-handler":
            # Legacy code. Need to rename handler_type to api_call_handler so that the function can be called directly.
            command_response = api_call_handler(pending_command.payload)
        elif pending_command.handler_type == FILE_HANDLER:
            command_response = file_handler(pending_command.command_type, pending_command.payload)
        else:
            # Generic code to handle all Cluster commands
            command_response = globals()[pending_command.handler_type](pending_command.command_type, pending_command.payload)
        if command_response["exit_code"] == 0:
            if pending_command.status == CSConstants.CS_COMMAND_TODO_SYNC:
                cluster.mark_command_complete_new(pending_command.id, command_response)
            else:
                cluster.dismiss_command(pending_command.id)
        else:
            logger.warning('Command failed with error: ' + str(command_response))
            cluster.mark_command_failed_new(pending_command.id, command_response)

    except Exception as e:
        logger.exception('Error executing command: %s' % pending_command)
        command_response = build_cluster_command_response(exit_code=1, status_code=500, error=str(e))
        cluster.mark_command_failed_new(pending_command.id, command_response)

def run_cluster_command(pending_command):
    commandId = pending_command.id
    command = pending_command.command
    status = pending_command.status
    try:
        output = None
        error = None
        returncode = None
        cluster.mark_command_in_progress(commandId)

        if pending_command.handler_type is not None:
            run_cluster_command_new(pending_command)
            return

        # Keeping below code for backward compatibility

        if command:
            if command.startswith('regenerate-appliance-keys'):
                returncode, output, error = handle_regenerate_appliance_keys_command(command.split(' ')[1])
            else:
                if not check_if_supported_cmd(command):
                    error = "Command [" + command + "] is not whitelisted in framework"
                    logger.warning('Command failed with error: ' + error)
                    cluster.mark_command_failed(commandId, error)
                    return
                command = shlex.split(command)
                logger.debug('Executing command ' + str(command)[:150])

                command_result = Popen(command, stdout=PIPE, stderr=PIPE)
                output, error = command_result.communicate()
                returncode = command_result.returncode

            if returncode == 0:
                if status == CSConstants.CS_COMMAND_TODO_SYNC:
                    cluster.mark_command_complete(commandId, output.decode('utf-8'))
                else:
                    cluster.dismiss_command(commandId)
            else:
                logger.warning('Command failed with error: '+error.decode('utf-8'))
                cluster.mark_command_failed(commandId, error.decode('utf-8'))
    except Exception as e:
        logger.exception('Error executing command: %s' % command)
        cluster.mark_command_failed(commandId, str(e))

if __name__ == '__main__':
    current_node = None
    # ensure that the node is not faulted
    current_node = get_node_from_db_or_list_node_file()
    while not current_node:
        # Case when the node has not yet joined self cluster
        logger.warning("Node is not part of any cluster")
        sleep(20)
        current_node = get_node_from_db_or_list_node_file()
    status = current_node['status']
    if status.lower() == 'faulted':
        logger.error('Node status is faulted. Wait for 2 mins for boot and disable all services.')
        sleep(120)
        serv_list = get_cyops_services_list()
        for serv in serv_list:
            # do not kill self
            if serv.strip() != 'cyops-ha':
                cmd = "systemctl stop " + serv
                execute_cmd(cmd)
                cmd = "systemctl disable " + serv
                execute_cmd(cmd)
                cmd = "systemctl mask " + serv
                execute_cmd(cmd)
        # only ha service keeps running forever
        while True:
            sleep(120)
    else:
        handler.schedule_health_check()
        while True:
            schedule.run_pending()
            try:
                pending_commands = cluster.get_self_cluster_command()
                for pending_command in pending_commands:
                    if pending_command.commandId == COMMAND_ID_REFRESH_CONFIG:
                        handler.refresh_health_check_schedule()
                        cluster.dismiss_command(pending_command.id)
                        continue

                    # Intially, we asssumed that all commands will be independent of each other and there will be no
                    # restriction of ordering. However, that is no longer valid as the commands to install solution pack
                    # need to be executed in order. For now, we are fixing it for message_handler only as
                    # message_handler is used to install solution pack. However, a complete fix will be to allow clients
                    # to execute commands in sequence. We will take that up in next release.
                    if pending_command.handler_type is not None and pending_command.handler_type == MESSAGE_HANDLER:
                        run_cluster_command(pending_command)
                        continue
                    # run the command in a separate thread
                    t = threading.Thread(target=run_cluster_command, args=[pending_command])
                    t.daemon = True
                    # TODO: validate that the thread exits once the command completes
                    t.start()
            except DBAPIError as db_error:
                logger.warning("HA service failed to connect to Database")
            except Exception as e:
                logger.exception("Error fetching cluster command")
            # sleep for 10 seconds
            sleep(10)
            try:
                cleanup_shared_data()
            except Exception as e:
                logger.warning("Error while cleaning up shared data")