#!/opt/cyops-auth/.env/bin/python
""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import os
import sys

sys.path.append("/opt/cyops-auth")


import argparse
from time import sleep
from traceback import print_exception, print_exc


if not os.geteuid() == 0:
    sys.exit("You need to be root to perform this command.")

fields_to_show = []
fields_to_show.append({"label": "Type           :", "field_name": "type"})
fields_to_show.append({"label": "Edition        :", "field_name": "edition"})
fields_to_show.append({"label": "Customer Name  :", "field_name": "customer_name"})
fields_to_show.append({"label": "Role           :", "field_name": "role"})
fields_to_show.append({"label": "Total Users    :", "field_name": "total_users"})
fields_to_show.append({"label": "Expiry Date    :", "field_name": "expiring_on"})
fields_to_show.append({"label": "Remaining Days :", "field_name": "remaining_days"})
fields_to_show.append({"label": "Serial no      :", "field_name": "serial_no"})

# Below fields will be there in license file when user deploy license
fields_to_show.append({"label": "Days           :", "field_name": "days"})
fields_to_show.append({"label": "Total Users    :", "field_name": "max_users"})

fields_to_skip_for_perpetual_and_extension_license = ['expiring_on', 'remaining_days', 'days']
fields_to_skip_for_subscription_license = ['max_users', 'days']


# Include common packages in licensemanager
sys.path.append("/opt/cyops/python_packages/fortisoar")
from handlerworkers.license import CSLicenseManager, NODE_HARDWARE_KEY, is_fsr_running_as_extension
from handlerworkers.fdnclient import fdn_sync, validate_license_fdn_part, reset_fdn_sync_scheduler
from utilities.csconstants import *
from utilities.ha.common_utils import execute_cmd


def show_license_details(license_details, from_file=False):
    for field in fields_to_show:
        if field['field_name'] in license_details.keys():
            if license_details['type'] in [PERPETUAL_TYPE, TRIAL_EXTENSION_TYPE] and field[
                'field_name'] in fields_to_skip_for_perpetual_and_extension_license:
                continue
            if from_file and license_details['type'] == SUBSCRIPTION_TYPE and field[
                'field_name'] in fields_to_skip_for_subscription_license:
                continue
            if license_details[field['field_name']] is not None:
                # Special handling is required for field 'role'
                if field['field_name'] == "role" and len(license_details[field['field_name']]) > 0:
                    print(field['label'] + " " + str(license_details[field['field_name']][0]))
                else:
                    # If subtype there, show it in ()
                    field_contents = str(license_details[field['field_name']])
                    if field['field_name'] == LICENSE_TYPE:
                        if (LICENSE_SUBTYPE in license_details.keys()) and license_details[LICENSE_SUBTYPE] is not None:
                            field_contents = field_contents + "(" + str(license_details[LICENSE_SUBTYPE]) + ")"
                    print(field['label'] + " " + field_contents)
    if from_file:
        print("Device UUID    : " + license_details['hardware_key'])
    else:
        print("Device UUID    : " + NODE_HARDWARE_KEY)
    return


def show_installed_license_details():
    try:
        if not CSLicenseManager().is_license_installed(NODE_HARDWARE_KEY):
            print("License not found on the system.")
            exit(1)
    except Exception as e:
        print("License not found on the system.")
        exit(1)
    is_valid_license, message = CSLicenseManager().validate_license()
    if not is_valid_license:
        print(message)
        if DUPLICATE_LICENSE_THRESHOLD_EXCEED_MSG_ERROR_CODE not in message:
            exit(1)
    license_info = CSLicenseManager().get_license_info()
    current_node_details = license_info['nodes'][NODE_HARDWARE_KEY]['details']
    show_license_details(current_node_details)
    return


def show_license_details_from_license_file(license_file):
    with open(license_file) as license_file:
        license_key = license_file.read()
    is_valid, new_license_claims = CSLicenseManager().verify_license_signature(license_key)
    if not is_valid:
        message = "License signature verification failed."
        print(message)
        exit(1)
    show_license_details(new_license_claims, from_file=True)


def take_arguments():
    print("1. Get Device UUID")
    print("2. Deploy License File")
    print("Press Ctrl+C to exit")

    option = input("Select an option: ")
    if not option:
        print("Please provide an input")
        take_arguments()
    else:
        try:
            option = int(option)
            if option != 1 and option != 2:
                print(option)
                print("Wrong option selected. Please Try again")
                sys.exit()
        except ValueError as error:
            print("Invalid input argument provided. Please try again\n")
            take_arguments()

    if option == 1:
        print("Device UUID: " + NODE_HARDWARE_KEY)

    elif option == 2:
        while True:
            option = input("Is the license issued for distributed multi-tenancy setup(y/n/yes/no)? ").lower().strip()
            if option == 'y' or option == 'yes':
                is_distributed = True
                break
            elif option == 'n' or option == 'no':
                is_distributed = False
                break
            else:
                print("Please provide valid input(y/n/yes/no)")
        file_path = input("License file path: ")
        deploy_license_from_file(NODE_HARDWARE_KEY, file_path, is_distributed)


def is_valid_file(file):
    if not os.path.exists(file):
        raise argparse.ArgumentTypeError("File {0} does not exist".format(file))
    return file


def start():
    parser = argparse.ArgumentParser()
    parser.prog = "csadm license"
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--get-hkey', dest='get_hkey', action='store_true', help=argparse.SUPPRESS)
    group.add_argument('--get-device-uuid', dest='get_hkey', action='store_true',
                       help='Retrieve device uuid for this instance')
    group.add_argument('--deploy-enterprise-license', dest="enterprise_license", help='Deploy enterprise license', type=is_valid_file)
    group.add_argument('--deploy-multi-tenant-license', dest="distributed_license", help='Deploy multi-tenant license', type=is_valid_file)
    parser.add_argument('--nodeId', dest='node_id', help=argparse.SUPPRESS)
    group.add_argument('--show-details', nargs='?', action='store', type=str, dest='license_file_path',
                       help='Show license details', required=False, default=None, const="dummy_value")
    group.add_argument('--deploy-extension-license', dest="extension_license", action='store_true',
                       help=argparse.SUPPRESS)
    group.add_argument('--sync-license', dest="sync_license", action='store_true',
                       help=argparse.SUPPRESS)
    group.add_argument('--get-device-sn', dest='get_device_sn', action='store_true',
                       help='Retrieve device SN for this instance')
    group.add_argument('--refresh-device-uuid', dest='refresh_device_uuid', action='store_true',
                       help='Refresh device uuid if the device uuid is changed')
    group.add_argument('--restore-license', dest='restore_license', action='store_true',
                       help=argparse.SUPPRESS)

    args = parser.parse_args()

    if args.get_hkey:
        print(NODE_HARDWARE_KEY)
        return
    if args.refresh_device_uuid:
        (status, message) = CSLicenseManager().refresh_device_uuid()
        if message:
            print(message)
        return
    if args.get_device_sn:
        serial_no = CSLicenseManager().get_device_sn()
        if serial_no:
            print(serial_no)
        else:
            print(LICENSE_NOT_FOUND_MSG)
        return

    if args.distributed_license:
        deploy_license_from_file(args.node_id, args.distributed_license, True)
        return
    if args.enterprise_license:
        deploy_license_from_file(args.node_id, args.enterprise_license, False)
        return
    if args.restore_license:
        license_obj = CSLicenseManager().get_license_objects(SELF_NODE_ID)
        if not license_obj or not license_obj.license_key:
            print("No license available to restore. Exiting")
            return
        license_key = license_obj.license_key
        unverified_claims = CSLicenseManager().get_unverified_claims(license_key)
        is_distributed = unverified_claims['edition'] != 'Enterprise'
        CSLicenseManager().delete_node_license()
        deploy_license(SELF_NODE_ID, license_key, is_distributed, license_obj.eula_accepted)
        return
    if args.sync_license:
        sync_license_details()
        return
    if args.extension_license:
        extension_license = CSLicenseManager().get_extension_license()
        if not extension_license:
            print(format_message_for_cli('FortiSOAR not running as FMG extension. Exiting..'))
            sys.exit(1)
        deploy_license(NODE_HARDWARE_KEY, extension_license, is_distributed=False)
        return
    if args.license_file_path is not None:
        if args.license_file_path == "dummy_value":
            # File is not provided as input
            show_installed_license_details()
        else:
            is_valid_file(args.license_file_path)
            show_license_details_from_license_file(args.license_file_path)
        return
    print("Command to either generate device uuid or to deploy a license\n")
    try:
        take_arguments()
    except KeyboardInterrupt:
        print("")


def validate_license_before_deployment(node_id, license_key, is_distributed):
    # validate inline part
    contracts = {}
    is_valid, message = CSLicenseManager().validate_license_inline_part(node_id, license_key, is_distributed,
                                                 is_cli_deploy=True)
    if not is_valid:
        return is_valid, message, contracts
    # Validate FDN part
    is_valid, message, contracts = validate_license_fdn_part(node_id, license_key)

    # For trial license, since the license generation and license deployment happens almost immediately, the license
    # details may not be present on FDN. Therefore, we will skip license validation against FDN, if the details are not
    # found on FDN
    LICENSE_NOT_REGISTERED_SUBSTRING = "is not registered with licensing server"
    if not is_valid and CSLicenseManager().is_trial_license(license_key) and LICENSE_NOT_REGISTERED_SUBSTRING in message:
        return True, "", contracts
    else:
        return is_valid, message, contracts


def deploy_license(nodeId, license_key, is_distributed, accept_eula=False):
    exit_code=1
    warning_message = None
    print(format_message_for_cli(INITIAL_MSG))
    print(format_message_for_cli(VALIDATING_LICENSE_MSG))
    if not nodeId:
        nodeId = NODE_HARDWARE_KEY
    try:
        # If the device uuid of the system is changed, refresh the device uuid
        CSLicenseManager().refresh_device_uuid()
        CSLicenseManager().mark_license_deployment_in_progress(nodeId)
        is_valid, message, contracts = validate_license_before_deployment(nodeId, license_key, is_distributed)
        if is_valid:
            if message:
                warning_message = message
            deployment_status, message = CSLicenseManager().deploy_license(nodeId, license_key, contracts, warning_message)
            is_valid = deployment_status
            if deployment_status is True:
                if message == LICENSE_DEPLOYMENT_SUCCESS_MSG:
                    if accept_eula:
                        CSLicenseManager().accept_eula()
                    # Reconfigure syslog so that it picks updated serial number
                    execute_cmd('csadm log forward refresh-config', ignore_failure=True, is_blocking=False, use_shell=False)
                    if is_distributed:
                        # Wait for postman and agent service to come up.
                        sleep(10)
                exit_code = 0
        if not is_valid:
            CSLicenseManager().mark_license_deployment_failed(nodeId, message)
        if warning_message and message == LICENSE_DEPLOYMENT_SUCCESS_MSG:
            message = LICENSE_DEPLOYMENT_WITH_WARNING.format(warning_message)
            is_warning = True

            # Do not print duplicate warning message for FSR MEA
            if not (is_fsr_running_as_extension and DUPLICATE_LICENSE_ERROR_CODE in message):
                print(format_message_for_cli(message, is_warning))
        else:
            print(message)
    except Exception as error:
        print_exc()
        print("License deployment failed with error: "+str(error))
        CSLicenseManager().mark_license_deployment_failed(nodeId, str(error))
    # This exit code is needed for "licensemanager command triggered during license deployment from UI".
    sys.exit(exit_code)


def deploy_license_from_file(nodeId, license_file, is_distributed):
    try:
        with open(license_file) as license_file:
            license_key = license_file.read()
        deploy_license(nodeId, license_key, is_distributed)
    except FileNotFoundError as error:
        print(f"The file {license_file} not found. Please try again")

    # Remove temporary license file created during ui license deployment
    if license_file == TEMP_LICENSE_PATH:
        os.remove(TEMP_LICENSE_PATH)


def sync_license_details():
    fdn_sync()


if __name__ == "__main__":
    start()